using System;
using Microsoft.Win32.SafeHandles;
using System.Diagnostics;
using System.Windows.Forms;

namespace WinUsbDemo
{
	/// <summary>
	///  Routines for the WinUsb driver.
	///  </summary>
	///  <remarks>
	///  I could find no way to test for a valid winUsbHandle. If the handle is 
	///  already closed, the application gets an AccessViolationException, which the 
	///  routines capture and ignore.
	/// </remarks>
	
	internal sealed partial class WinUsbCommunications
	{
		/// <summary>
		/// Stores information about the WinUSB device.
		/// </summary>

		private const String ModuleName = "WinUSB Device";

		public class DeviceInfo
		{
			internal Byte BulkInPipe;
			internal Byte BulkOutPipe;
			internal Byte InterruptInPipe;
			internal Byte InterruptOutPipe;
			internal UInt32 DeviceSpeed;
		}

		///  <summary>
		///  Closes the device handle obtained with CreateFile and frees resources.
		///  </summary>
		///  <param name="deviceHandle"> Device handle obtained with CreateFile </param>
		///  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>

		internal void CloseDeviceHandle(SafeFileHandle deviceHandle, IntPtr winUsbHandle)
		{
			try
			{
				NativeMethods.WinUsb_Free(winUsbHandle);

				if (deviceHandle != null)
				{
					if (!(deviceHandle.IsInvalid))
					{
						deviceHandle.Close();
					}
				}
			}
			catch (AccessViolationException ex)
			{
				// The handle might be closed.
			}

			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Provides a central mechanism for exception handling.
		///  Displays a message box that describes the exception.
		///  </summary>
		///  
		///  <param name="Name"> the module where the exception occurred. </param>
		///  <param name="e"> the exception </param>

		internal static void DisplayException(String Name, Exception e)
		{
			try
			{
				//  Create an error message.

				String message = "Exception: " + e.Message + Environment.NewLine + "Module: " + Name + Environment.NewLine + "Method: " +
				          e.TargetSite.Name;

				const String caption = "Unexpected Exception";

				MessageBox.Show(message, caption, MessageBoxButtons.OK);
				Debug.Write(message);
			}
			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Initiates a Control Read transfer. Data stage is device to host.
		///  </summary>
		/// 
		///  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		///  <param name="dataStage"> The received data </param>
		///  
		///  <returns>
		///  True on success, False on failure.
		///  </returns>

		internal Boolean DoControlReadTransfer(IntPtr winUsbHandle, ref Byte[] dataStage)
		{
			UInt32 bytesReturned = 0;

			try
			{
				//  Vendor-specific request to an interface with device-to-host Data stage.

				NativeMethods.WINUSB_SETUP_PACKET setupPacket;
				setupPacket.RequestType = 0XC1;

				//  The request number that identifies the specific request.

				setupPacket.Request = 2;

				//  Command-specific value to send to the device.
				// For a request directed to an interface, the WinUSB driver uses this field to specify the interface number
				// and will ignore a value set here.
				// For a request directed to the device, you can use this field for vendor-defined purposes.

				setupPacket.Index = 0;

				//  Number of bytes in the request's Data stage.

				setupPacket.Length = Convert.ToUInt16(dataStage.Length);

				//  Command-specific value to send to the device.

				setupPacket.Value = 0;

				// ***
				//  winusb function 

				//  summary
				//  Initiates a control transfer.

				//  paramaters
				//  Device handle returned by WinUsb_Initialize.
				//  WINUSB_SETUP_PACKET structure 
				//  Buffer to hold the returned Data-stage data.
				//  Number of data bytes to read in the Data stage.
				//  Number of bytes read in the Data stage.
				//  Null pointer for non-overlapped.

				//  returns
				//  True on success.
				//  ***            

				var success = NativeMethods.WinUsb_ControlTransfer(winUsbHandle, setupPacket, dataStage,
				                                                   Convert.ToUInt16(dataStage.Length), ref bytesReturned,
				                                                   IntPtr.Zero);
				return success;

			}
			catch (AccessViolationException ex)
			{
				// The handle might be closed.

				return false;
			}
			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Initiates a Control Write transfer. Data stage is host to device.
		///  </summary>
		///  
		///  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param> 
		///  <param name="dataStage"> The data to send. </param>
		///  
		///  <returns>
		///  True on success, False on failure.
		///  </returns>

		internal Boolean DoControlWriteTransfer(IntPtr winUsbHandle, Byte[] dataStage)
		{
			uint bytesReturned = 0;
			var value = Convert.ToUInt16(0);

			try
			{
				//  Vendor-specific request to an interface with host-to-device Data stage.

				NativeMethods.WINUSB_SETUP_PACKET setupPacket;
				setupPacket.RequestType = 0X41;

				//  The request number that identifies the specific request.

				setupPacket.Request = 1;

				//  Command-specific value to send to the device.
				// For a request directed to an interface, the WinUSB driver uses this field to specify the interface number.
				// and will ignore a value set here.
				// For a request directed to the device, you can use this field for vendor-defined purposes.

				setupPacket.Index = 0;

				//  Number of bytes in the request's Data stage.

				setupPacket.Length = Convert.ToUInt16(dataStage.Length);

				//  Command-specific value to send to the device.

				setupPacket.Value = value;

				// ***
				//  winusb function 

				//  summary
				//  Initiates a control transfer.

				//  parameters
				//  Device handle returned by WinUsb_Initialize.
				//  WINUSB_SETUP_PACKET structure 
				//  Buffer containing the Data-stage data.
				//  Number of data bytes to send in the Data stage.
				//  Number of bytes sent in the Data stage.
				//  Null pointer for non-overlapped.

				//  Returns
				//  True on success.
				//  ***

				var success = NativeMethods.WinUsb_ControlTransfer
					(winUsbHandle,
					 setupPacket,
					 dataStage,
					 Convert.ToUInt16(dataStage.Length),
					 ref bytesReturned,
					 IntPtr.Zero);
				return success;
			}
			catch (AccessViolationException ex)
			{
				// The handle might be closed.
				return false;
			}
			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Requests a handle with CreateFile.
		///  </summary>
		///  
		///  <param name="devicePathName"> Returned by SetupDiGetDeviceInterfaceDetail 
		///  in an SP_DEVICE_INTERFACE_DETAIL_DATA structure. </param>
		///  
		///  <returns>
		///  The handle.
		///  </returns>

		internal SafeFileHandle GetDeviceHandle(String devicePathName)
		{
			try
			{
				// ***
				// API function

				//  summary
				//  Retrieves a handle to a device.

				//  parameters 
				//  Device path name returned by SetupDiGetDeviceInterfaceDetail
				//  Type of access requested (read/write).
				//  FILE_SHARE attributes to allow other processes to access the device while this handle is open.
				//  Security structure. Using Null for this may cause problems under Windows XP.
				//  Creation disposition value. Use OPEN_EXISTING for devices.
				//  Flags and attributes for files. The winsub driver requires FILE_FLAG_OVERLAPPED.
				//  Handle to a template file. Not used.

				//  Returns
				//  A handle or INVALID_HANDLE_VALUE.
				// ***

				SafeFileHandle deviceHandle = FileIo.NativeMethods.CreateFile
					(devicePathName,
					 (FileIo.NativeMethods.GENERIC_WRITE | FileIo.NativeMethods.GENERIC_READ),
					 FileIo.NativeMethods.FILE_SHARE_READ | FileIo.NativeMethods.FILE_SHARE_WRITE,
					 IntPtr.Zero,
					 FileIo.NativeMethods.OPEN_EXISTING,
					 FileIo.NativeMethods.FILE_ATTRIBUTE_NORMAL | FileIo.NativeMethods.FILE_FLAG_OVERLAPPED,
					 IntPtr.Zero);
				return deviceHandle;
			}
			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Initializes a device interface and obtains information about it.
		///  Calls these winusb API functions:
		///    WinUsb_Initialize
		///    WinUsb_QueryInterfaceSettings
		///    WinUsb_QueryPipe
		///  </summary>
		///  
		///  <param name="deviceHandle"> Device handle obtained with CreateFile </param>
		///  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		///  <param name="myDeviceInfo"> devInfo structure for the device </param>	
		///  <param name="pipeTimeout"> desired timeout in milliseconds for transfers </param>
		///  
		///  <returns>
		///  True on success, False on failure.
		///  </returns>

		internal Boolean InitializeDevice(SafeFileHandle deviceHandle, ref IntPtr winUsbHandle, ref DeviceInfo myDeviceInfo,
		                                  UInt32 pipeTimeout)
		{
			try
			{
				NativeMethods.USB_INTERFACE_DESCRIPTOR ifaceDescriptor;
				ifaceDescriptor.bLength = 0;
				ifaceDescriptor.bDescriptorType = 0;
				ifaceDescriptor.bInterfaceNumber = 0;
				ifaceDescriptor.bAlternateSetting = 0;
				ifaceDescriptor.bNumEndpoints = 0;
				ifaceDescriptor.bInterfaceClass = 0;
				ifaceDescriptor.bInterfaceSubClass = 0;
				ifaceDescriptor.bInterfaceProtocol = 0;
				ifaceDescriptor.iInterface = 0;

				NativeMethods.WINUSB_PIPE_INFORMATION pipeInfo;
				pipeInfo.PipeType = 0;
				pipeInfo.PipeId = 0;
				pipeInfo.MaximumPacketSize = 0;
				pipeInfo.Interval = 0;

				// ***
				//  winusb function 

				//  summary
				//  get a handle for communications with a winusb device        '

				//  parameters
				//  Handle returned by CreateFile.
				//  Device handle to be returned.

				//  returns
				//  True on success.
				//  ***

				var success = NativeMethods.WinUsb_Initialize
					(deviceHandle,
					 ref winUsbHandle);

				if (success)
				{
					// ***
					//  winusb function 

					//  summary
					//  Get a structure with information about the device interface.

					//  parameters
					//  handle returned by WinUsb_Initialize
					//  alternate interface setting number
					//  USB_INTERFACE_DESCRIPTOR structure to be returned.

					//  returns
					//  True on success.

					success = NativeMethods.WinUsb_QueryInterfaceSettings
						(winUsbHandle,
						 0,
						 ref ifaceDescriptor);

					if (success)
					{
						//  Get the transfer type, endpoint number, and direction for the interface's
						//  bulk and interrupt endpoints. Set pipe policies.

						// ***
						//  winusb function 

						//  summary
						//  returns information about a USB pipe (endpoint address)

						//  parameters
						//  Handle returned by WinUsb_Initialize
						//  Alternate interface setting number
						//  Number of an endpoint address associated with the interface. 
						//  (The values count up from zero and are NOT the same as the endpoint address
						//  in the endpoint descriptor.)
						//  WINUSB_PIPE_INFORMATION structure to be returned

						//  returns
						//  True on success   
						// ***

						for (var i = 0; i <= ifaceDescriptor.bNumEndpoints - 1; i++)
						{
							NativeMethods.WinUsb_QueryPipe
								(winUsbHandle,
								 0,
								 Convert.ToByte(i),
								 ref pipeInfo);

							if (((pipeInfo.PipeType ==
							      NativeMethods.USBD_PIPE_TYPE.UsbdPipeTypeBulk) &
							     UsbEndpointDirectionIn(pipeInfo.PipeId)))
							{
								myDeviceInfo.BulkInPipe = pipeInfo.PipeId;

								SetPipePolicy
									(winUsbHandle,
									 myDeviceInfo.BulkInPipe,
									 Convert.ToUInt32(NativeMethods.POLICY_TYPE.IGNORE_SHORT_PACKETS),
									 Convert.ToByte(false));

								SetPipePolicy
									(winUsbHandle,
									 myDeviceInfo.BulkInPipe,
									 Convert.ToUInt32(NativeMethods.POLICY_TYPE.PIPE_TRANSFER_TIMEOUT),
									 pipeTimeout);
							}
							else if (((pipeInfo.PipeType ==
							           NativeMethods.USBD_PIPE_TYPE.UsbdPipeTypeBulk) &
							          UsbEndpointDirectionOut(pipeInfo.PipeId)))
							{
								myDeviceInfo.BulkOutPipe = pipeInfo.PipeId;

								SetPipePolicy
									(winUsbHandle,
									 myDeviceInfo.BulkOutPipe,
									 Convert.ToUInt32(NativeMethods.POLICY_TYPE.IGNORE_SHORT_PACKETS),
									 Convert.ToByte(false));

								SetPipePolicy
									(winUsbHandle,
									 myDeviceInfo.BulkOutPipe,
									 Convert.ToUInt32(NativeMethods.POLICY_TYPE.PIPE_TRANSFER_TIMEOUT),
									 pipeTimeout);
							}
							else if ((pipeInfo.PipeType ==
							          NativeMethods.USBD_PIPE_TYPE.UsbdPipeTypeInterrupt) &
							         UsbEndpointDirectionIn(pipeInfo.PipeId))
							{
								myDeviceInfo.InterruptInPipe = pipeInfo.PipeId;

								SetPipePolicy
									(winUsbHandle,
									 myDeviceInfo.InterruptInPipe,
									 Convert.ToUInt32(NativeMethods.POLICY_TYPE.IGNORE_SHORT_PACKETS),
									 Convert.ToByte(false));

								SetPipePolicy
									(winUsbHandle,
									 myDeviceInfo.InterruptInPipe,
									 Convert.ToUInt32(NativeMethods.POLICY_TYPE.PIPE_TRANSFER_TIMEOUT),
									 pipeTimeout);
							}
							else if ((pipeInfo.PipeType ==
							          NativeMethods.USBD_PIPE_TYPE.UsbdPipeTypeInterrupt) &
							         UsbEndpointDirectionOut(pipeInfo.PipeId))
							{
								myDeviceInfo.InterruptOutPipe = pipeInfo.PipeId;

								SetPipePolicy
									(winUsbHandle,
									 myDeviceInfo.InterruptOutPipe,
									 Convert.ToUInt32(NativeMethods.POLICY_TYPE.IGNORE_SHORT_PACKETS),
									 Convert.ToByte(false));

								SetPipePolicy
									(winUsbHandle,
									 myDeviceInfo.InterruptOutPipe,
									 Convert.ToUInt32(NativeMethods.POLICY_TYPE.PIPE_TRANSFER_TIMEOUT),
									 pipeTimeout);
							}
						}
					}
				}
				return success;
			}

			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Is the current operating system Windows XP or later?
		///  The WinUSB driver requires Windows XP or later.
		///  </summary>
		/// 
		///  <returns>
		///  True if Windows XP or later, False if not.
		///  </returns>

		internal Boolean IsWindowsXpOrLater()
		{
			try
			{
				var myEnvironment = Environment.OSVersion;

				//  Windows XP is version 5.1.

				var versionXp = new Version(5, 1);

				if (myEnvironment.Version >= versionXp)
				{
					return true;
				}
				return false;
			}

			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Gets a value that corresponds to a USB_DEVICE_SPEED. 
		///  </summary>
		///  
		///  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		///  <param name="myDevInfo"> devInfo structure for the device </param>

		internal static Boolean QueryDeviceSpeed(IntPtr winUsbHandle, ref DeviceInfo myDevInfo)
		{
			UInt32 length = 1;
			var speed = new Byte[1];

			try
			{
				// ***
				//  winusb function 

				//  summary
				//  Get the device speed. 
				//  (Normally not required but can be nice to know.)

				//  parameters
				//  Handle returned by WinUsb_Initialize
				//  Requested information type.
				//  Number of bytes to read.
				//  Information to be returned.

				//  returns
				//  True on success.
				// ***           			

				var success = NativeMethods.WinUsb_QueryDeviceInformation
					(winUsbHandle,
					 NativeMethods.DEVICE_SPEED,
					 ref length,
					 ref speed[0]);

				if (success)
				{
					myDevInfo.DeviceSpeed = Convert.ToUInt32(speed[0]);
				}

				return success;
			}
			catch (AccessViolationException ex)
			{
				// The handle might be closed.
				return false;
			}

			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Attempts to read data from a bulk IN endpoint.
		///  </summary>
		///  
		///  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		///  <param name="myDeviceInfo"> devInfo structure for the device </param>		
		///  <param name="bytesToRead"> Number of bytes to read. </param>
		///  <param name="dataBuffer"> Buffer for storing the bytes read. </param>
		///  <param name="bytesRead"> Number of bytes read. </param>
		///  <param name="success"> Success or failure status. </param>
		///  
		internal void ReceiveDataViaBulkTransfer(IntPtr winUsbHandle, DeviceInfo myDeviceInfo, UInt32 bytesToRead,
		                                         ref Byte[] dataBuffer, ref UInt32 bytesRead, ref Boolean success)
		{
			try
			{
				// ***
				//  winusb function 

				//  summary
				//  Attempts to read data from a device interface.

				//  parameters
				//  Device handle returned by WinUsb_Initialize.
				//  Endpoint address.
				//  Buffer to store the data.
				//  Maximum number of bytes to return.
				//  Number of bytes read.
				//  Null pointer for non-overlapped.

				//  Returns
				//  True on success.
				// ***

				{
					success = NativeMethods.WinUsb_ReadPipe
						(winUsbHandle,
						 myDeviceInfo.BulkInPipe,
						 dataBuffer,
						 bytesToRead,
						 ref bytesRead,
						 IntPtr.Zero);
				}
			}
			catch (AccessViolationException ex)
			{
				// The handle might be closed.
			}

			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Attempts to read data from an interrupt IN endpoint. 
		///  </summary>
		///  
		///  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		///  <param name="myDeviceInfo"> devInfo structure for the device </param>		
		///  <param name="bytesToRead"> Number of bytes to read. </param>
		///  <param name="dataBuffer"> Buffer for storing the bytes read. </param>
		///  <param name="bytesRead"> Number of bytes read. </param>
		///  
		///  <returns> true on success, false on failure </returns>
		///  
		internal Boolean ReceiveDataViaInterruptTransfer(IntPtr winUsbHandle, DeviceInfo myDeviceInfo, UInt32 bytesToRead,
		                                                 ref Byte[] dataBuffer, ref UInt32 bytesRead)
		{
			try
			{
				// ***
				//  winusb function 

				//  summary
				//  Attempts to read data from a device interface.

				//  parameters
				//  Device handle returned by WinUsb_Initialize.
				//  Endpoint address.
				//  Buffer to store the data.
				//  Maximum number of bytes to return.
				//  Number of bytes read.
				//  Null pointer for non-overlapped.

				//  Returns
				//  True on success.
				// ***

				Boolean success = NativeMethods.WinUsb_ReadPipe
					(winUsbHandle,
					 myDeviceInfo.InterruptInPipe,
					 dataBuffer,
					 bytesToRead,
					 ref bytesRead,
					 IntPtr.Zero);

				return success;
			}
			catch (AccessViolationException ex)
			{
				// The handle might be closed.
				return false;
			}


			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Attempts to send data via a bulk OUT endpoint.
		///  </summary>
		///  
		///  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		///  <param name="myDeviceInfo"> devInfo structure for the device </param> 		
		///  <param name="bytesToWrite"> Number of bytes to write. </param>
		///  <param name="dataBuffer"> Buffer containing the bytes to write. </param>
		///  <param name="bytesWritten"> The number of bytes written </param>
		///  <param name="success"> True on success </param>

		internal void SendDataViaBulkTransfer(IntPtr winUsbHandle, DeviceInfo myDeviceInfo, UInt32 bytesToWrite,
		                                      Byte[] dataBuffer, ref UInt32 bytesWritten, ref Boolean success)
		{
			try
			{
				// ***
				//  winusb function 

				//  summary
				//  Attempts to write data to a device interface.

				//  parameters
				//  Device handle returned by WinUsb_Initialize.
				//  Endpoint address.
				//  Buffer with data to write.
				//  Number of bytes to write.
				//  Number of bytes written.
				//  IntPtr.Zero for non-overlapped I/O.

				//  Returns
				//  True on success.
				//  ***

				{
					success = NativeMethods.WinUsb_WritePipe
						(winUsbHandle,
						 myDeviceInfo.BulkOutPipe,
						 dataBuffer,
						 bytesToWrite,
						 ref bytesWritten,
						 IntPtr.Zero);
				}
			}
			catch (AccessViolationException ex)
			{
				// The handle might be closed.
			}


			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Attempts to send data via an interrupt OUT endpoint.
		///  </summary>
		///
		///  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>  
		///  <param name="myDeviceInfo"> devInfo structure for the device </param> 		
		///  <param name="bytesToWrite"> Number of bytes to write. </param>
		///  <param name="dataBuffer"> Buffer containing the bytes to write. </param>
		///  <param name="bytesWritten"> The number of bytes written </param>		
		///  
		///  <returns>
		///  True on success, False on failure.
		///  </returns>

		internal Boolean SendDataViaInterruptTransfer(IntPtr winUsbHandle, DeviceInfo myDeviceInfo, UInt32 bytesToWrite,
		                                              Byte[] dataBuffer, ref UInt32 bytesWritten)
		{
			try
			{
				// ***
				//  winusb function 

				//  summary
				//  Attempts to write data to a device interface.

				//  parameters
				//  Device handle returned by WinUsb_Initialize.
				//  Endpoint address.
				//  Buffer with data to write.
				//  Number of bytes to write.
				//  Number of bytes written.
				//  IntPtr.Zero for non-overlapped I/O.

				//  Returns
				//  True on success.
				//  ***

				var success = NativeMethods.WinUsb_WritePipe
					(winUsbHandle,
					 myDeviceInfo.InterruptOutPipe,
					 dataBuffer,
					 bytesToWrite,
					 ref bytesWritten,
					 IntPtr.Zero);

				return success;
			}
			catch (AccessViolationException ex)
			{
				// The handle might be closed.
				return false;
			}


			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Sets pipe policy.
		///  Used when the value parameter is a Byte (all except PIPE_TRANSFER_TIMEOUT).
		///  </summary>
		/// 
		///  <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		/// <param name="pipeId"> Pipe to set a policy for. </param>
		///  <param name="policyType"> POLICY_TYPE member. </param>
		///  <param name="value"> Policy value. </param>
		///  
		///  <returns>
		///  True on success, False on failure.
		///  </returns>
		///  
		private Boolean SetPipePolicy(IntPtr winUsbHandle, Byte pipeId, UInt32 policyType, Byte value)
		{
			try
			{
				// ***
				//  winusb function 

				//  summary
				//  sets a pipe policy 

				//  parameters
				//  handle returned by WinUsb_Initialize
				//  identifies the pipe
				//  POLICY_TYPE member.
				//  length of value in bytes
				//  value to set for the policy.

				//  returns
				//  True on success 
				// ***

				var success = NativeMethods.WinUsb_SetPipePolicy
					(winUsbHandle,
					 pipeId,
					 policyType,
					 1,
					 ref value);

				return success;
			}
			catch (AccessViolationException ex)
			{
				// The handle might be closed.

				return false;
			}


			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Sets pipe policy.
		///  Used when the value parameter is a UInt32 (PIPE_TRANSFER_TIMEOUT only).
		///  </summary>
		/// 
		/// <param name="winUsbHandle"> Handle for accessing the WinUSB device </param>
		/// <param name="pipeId"> Pipe to set a policy for. </param>
		///  <param name="policyType"> POLICY_TYPE member. </param>
		///  <param name="value"> Policy value. </param>
		///  
		///  <returns>
		///  True on success, False on failure.
		///  </returns>
		///  
		private Boolean SetPipePolicy(IntPtr winUsbHandle, Byte pipeId, UInt32 policyType, UInt32 value)
		{
			var success = false;
			try
			{
				// ***
				//  winusb function 

				//  summary
				//  sets a pipe policy 

				//  parameters
				//  handle returned by WinUsb_Initialize
				//  identifies the pipe
				//  POLICY_TYPE member.
				//  length of value in bytes
				//  value to set for the policy.

				//  returns
				//  True on success 
				// ***

				success = NativeMethods.WinUsb_SetPipePolicy1
					(winUsbHandle,
					 pipeId,
					 policyType,
					 4,
					 ref value);

				return success;
			}
			catch (AccessViolationException ex)
			{
				// Handle might be closed.

				return success;

			}
			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Is the endpoint's direction IN (device to host)?
		///  </summary>
		///  
		///  <param name="endpointAddress"> The endpoint address. </param>
		///  
		///  <returns>
		///  True if IN (device to host), False if OUT (host to device)
		///  </returns> 

		private Boolean UsbEndpointDirectionIn(Int32 endpointAddress)
		{
			try
			{

				var directionIn = false;

				if (((endpointAddress & 0X80) == 0X80))
				{
					directionIn = true;
				}
				return directionIn;
			}

			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

		///  <summary>
		///  Is the endpoint's direction OUT (host to device)?
		///  </summary>
		///  
		///  <param name="addr"> The endpoint address. </param>
		///  
		///  <returns>
		///  True if OUT (host to device, False if IN (device to host)
		///  </returns>

		private Boolean UsbEndpointDirectionOut(Int32 addr)
		{
			try
			{
				var directionOut = false;

				if (((addr & 0X80) == 0))
				{
					directionOut = true;
				}
				return directionOut;
			}


			catch (Exception ex)
			{
				DisplayException(ModuleName, ex);
				throw;
			}
		}

	}
}
